/*
   256color rotating textured cubes.
   Needs 1 or 2 .gif pictures in the same directory to work.

   Tested with Glide 2.42, Watcom DOS.
   v.2
   March 99: updated for Banshee, because printf/puts
   messes up the Banshee graphics.

   -- Benny <benny_land@geocities.com>
*/
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string.h>
#include <malloc.h>
#include <math.h>

#include <glide.h>

#define BANSHEE

#define YES		1
#define NO		0
#define ALL_FINE 	""

#define MAXFACE			6	// cube faces = 6
#define MAXPOL			4	// vertexes
#define MAXTEX			2	// texture count

GrHwConfiguration hwconfig;
short           GlideIsRunning = 0;

// startup preferences
short           inside = NO,
		twosided = NO,
		allsides = YES,
		show3D = YES;

// Screen
long            size_x = 640, size_y = 480;
float           midx, midy;

// Cube
float           side1 = 255, side2 = 255;

short           which_texture = 0;

// --------------------------------------------------------------- //

void            quit(char *t);
void            ReadPicTex(int count);

typedef struct {
    float           x, y, z;
    float           w;
    float           s, t;
    float           r, g, b;
    float           a;
} Vertex3D;

Vertex3D        Verts[4];

FxU32           evenodd = GR_MIPMAPLEVELMASK_EVEN;
GrVertex        vtx[MAXPOL];

// --------------------------------------------------------------- //

#include "math3d.h"		// matrix functions
#include "usualgl3.h"		// usual stuff
#include "getpic.h"		// read a 256c pic

// --------------------------------------------------------------- //

PIC             pics[MAXTEX];

// --------------------------------------------------------------- //

void            RenderFace(MAT3D & M, float Xp, float Yp)
{
    static          lasttex = -1;
    int             j;
    //
    //  Project four corners
    //
    j = 0;
    // upper left:              // (0,      0,      0) (in world space)
    vtx[j].x = 0;
    vtx[j].y = 0;
    vtx[j].z = 0;
    j++;
    // upper right:             // (height, 0,      0)
    vtx[j].x = Xp;
    vtx[j].y = 0;
    vtx[j].z = 0;
    j++;
    // upper left:              // (height, length, 0)
    vtx[j].x = Xp;
    vtx[j].y = Yp;
    vtx[j].z = 0;
    j++;
    // lower right:             // (0,      length, 0)
    vtx[j].x = 0;
    vtx[j].y = Yp;
    vtx[j].z = 0;
    j++;

    for (j = 0; j < MAXPOL; j++) {
	Translate_Point(M, vtx[j].x, vtx[j].y, vtx[j].z);
	vtx[j].oow = 1. / vtx[j].z;
	vtx[j].x = vtx[j].x * vtx[j].oow + midx;	// x (in screen space)

	vtx[j].y = vtx[j].y * vtx[j].oow + midy;	// y

	SNAP(vtx[j]);
    }

    // textured mapping:
    for (j = 0; j < MAXPOL; j++) {
	vtx[j].tmuvtx[0].sow = Verts[j].s * pics[which_texture].new_x * vtx[j].oow;
	vtx[j].tmuvtx[0].tow = Verts[j].t * pics[which_texture].new_y * vtx[j].oow;
    }

    if (lasttex != which_texture) {
	grTexSource(GR_TMU0,
		    pics[which_texture].TextureAddress,
		    evenodd,
		    &pics[which_texture].info);
	lasttex = which_texture;
    }
    grDrawPolygonVertexList(MAXPOL, vtx);
}

void            info()
{
#ifndef BANSHEE
    char            nfo[132];
    strcpy(nfo, "Show: ");

    // View the inside/outside of the cube
    if (inside)
	strcat(nfo, "inside [v]iew, ");
    else
	strcat(nfo, "outside [v]iew, ");
    // Invisible plane toggle
    if (twosided)
	strcat(nfo, "two-[s]ided, ");
    else
	strcat(nfo, "1-[s]ided, ");
    // Show some/all faces
    if (allsides)
	strcat(nfo, "all [f]aces, ");
    else
	strcat(nfo, "some [f]aces, ");

    printf("%s\n", nfo);
#endif
}

void            RenderAFace(int j, MAT3D & M)
{
    switch (j) {
    case 2:
    case 5:
	RenderFace(M, side2, side2);
	break;
    default:
	RenderFace(M, side1, side2);
	break;
    }
}

void            RenderCube(MAT3D Sid[])
{
    int             j;
    for (j = 0; j < MAXFACE / 2; j++)
	if (allsides && (twosided || IsVisible(Sid[j])))
	    RenderAFace(j, Sid[j]);

    for (; j < MAXFACE; j++)
	if ((twosided || IsVisible(Sid[j])))
	    RenderAFace(j, Sid[j]);
}

void            MainLoop()
{
    int             j = 0, k, done = 0, running = 1, l;
    MAT3D           S[MAXFACE], Cube[MAXTEX], Side[MAXTEX][MAXFACE], Temp;

    BuildCube(S);

#ifndef BANSHEE
    puts("\nSPACE stops the motion, TAB switches displays.");
    puts("\nCurrent options:");
    info();
#endif

    while (!done) {
	if (running) {
	    j++;
	    for (l = 0; l < MAXTEX; l++) {
		Cube[l] = Translate(-128 - l * 128, -100, -100);
		Temp = RotateXYZ(j * 0.01, j * 0.001, j * 0.007);
		Cube[l] = Temp * Cube[l];
		Temp = Translate(0, 0, sin(-j * 0.00121 * (l + 1) * 10) * (350 - l * 100) + 600);
		Cube[l] = Temp * Cube[l];
		Temp = ScaleXYZ(200, 200, 1);
		Cube[l] = Temp * Cube[l];
	    }
	}
	for (k = 0; k < MAXFACE; k++)
	    for (l = 0; l < MAXTEX; l++)
		Side[l][k] = Cube[l] * S[k];

	grBufferClear(0, 0, GR_WDEPTHVALUE_FARTHEST);

	for (l = 0; l < MAXTEX; l++) {
	    which_texture = l;
	    midx = size_x / 3 + size_x / 4 * l;		// off-center

	    RenderCube(Side[l]);
	}

	grBufferSwap(1);

	if (kbhit())
	    switch (getch()) {
	    case 0x1b:		// VK_ESCAPE:
		done = YES;
		break;
	    case 0xD:		// VK_RETURN:
	    case 's':
		twosided = !twosided;
		info();
		break;
	    case 0x8:		// VK_BACK:
	    case 'v':
		inside = !inside;
		info();
		break;
	    case ' ':		// VK_SPACE:
		running = !running;
		break;
	    case '+':
	    case 'f':
		allsides = !allsides;
		info();
		break;
#ifndef BANSHEE
	    case 9:		// VK_TAB:
		show3D = !show3D;
		if (show3D)
		    grSstControl(GR_CONTROL_ACTIVATE);	// 3dfx on
		else
		    grSstControl(GR_CONTROL_DEACTIVATE); // 3dfx off

		break;
#endif
	    }
	if (done)
	    break;
    }
}

void            SetupGlideStates()
{
    // Let's use the w depth buffer
    grDepthBufferMode(GR_DEPTHBUFFER_WBUFFER);
    grDepthMask(FXTRUE);

    // Let's use the textures
    grColorCombine(GR_COMBINE_FUNCTION_SCALE_OTHER,
		   GR_COMBINE_FACTOR_ONE,
		   GR_COMBINE_LOCAL_NONE,	// <-
		   GR_COMBINE_OTHER_TEXTURE,	// <-
		   FXFALSE);

    grTexCombine(GR_TMU0,
		 GR_COMBINE_FUNCTION_LOCAL,
		 GR_COMBINE_FACTOR_NONE,
		 GR_COMBINE_FUNCTION_LOCAL,
		 GR_COMBINE_FACTOR_NONE,
		 FXFALSE,
		 FXFALSE);
}

void            GenInit()
{
    int             j;

    StartGlide();
    ReadPicTex(MAXTEX);

    lowfpu();

    midx = size_x / 2;
    midy = size_y / 2;

#define VX0	       0
#define VX1	       1

    // Let's use the full extend of the texture
    Verts[0].s = VX1;
    Verts[0].t = VX0;
    Verts[1].s = VX0;
    Verts[1].t = VX0;
    Verts[2].s = VX0;
    Verts[2].t = VX1;
    Verts[3].s = VX1;
    Verts[3].t = VX1;
}

/*
   Only one palette allowed per TMU,
   so we must match subsequent palettes to the first one.
   (can be ugly if palette #1 has too few colors...)
 */
#define SquaredX(x)   ((long)(x)*(x))
uchar           ClosestColorMatch(uchar * OldPal, uchar r, uchar g, uchar b)
{
    long            d = 2000000L;	// current "distance" away from ideal
    long            dc;			// calculated d;
    short           closest = 0xFF, i;	// closest color so far

    for (i = 0; i < 256; i++) {
	dc = SquaredX((long) OldPal[i * 3 + 0] - r) +
	    SquaredX((long) OldPal[i * 3 + 1] - g) +
	    SquaredX((long) OldPal[i * 3 + 2] - b);
	if (dc < d) {
	    closest = i;
	    d = dc;
	}
    }
    return closest;
}

void            CalcLookup(uchar * lookup, uchar * OldPal, uchar * NewPal, short count)
{
    int             i;

    memset(lookup, 0, 256);
    for (i = 0; i < count; i++)
	lookup[i] = ClosestColorMatch(OldPal,
		   NewPal[i * 3 + 0], NewPal[i * 3 + 1], NewPal[i * 3 + 2]);
}

// Inspired from mdldemo.c by Trey
void            PicToTMU(PIC * g)
{
    FxU32           textureSize;
    GrAspectRatio_t asp;
    GrLOD_t         lod;
    GuTexPalette    palette;
    char           *g_info_data = (char *) g->info.data;
    int             i, j;
    int             tw, th;
    static int      paletteloaded = 0;
    static char     firstpal[768];
    uchar           lookup[256];
    uchar           pixel;

    if (!paletteloaded) {
	memset(&palette, 0, sizeof(GuTexPalette));
	memset(&firstpal, 0, 768);
	memcpy(&firstpal, g->palette, 3 * g->c);

	// 3dfx palette
	for (i = 0; i < g->c; i++) {
	    palette.data[i] = 0;
	    palette.data[i] |= (int) (g->palette[i * 3 + 0]) << 16;
	    palette.data[i] |= (int) (g->palette[i * 3 + 1]) << 8;
	    palette.data[i] |= (int) (g->palette[i * 3 + 2]);
	}

	// Download the palette
	grTexDownloadTable(GR_TMU0, GR_TEXTABLE_PALETTE, &palette);

	for (j = 0; j < 256; j++)
	    lookup[j] = j;

	paletteloaded = 1;

    } else {
	// we could do a memcmp instead, to see if the palettes are equal...
	CalcLookup((uchar *) & lookup,
		   (uchar *) & firstpal, (uchar *) g->palette,
		   g->c);
    }

    // Texture size
    tw = g->x;
    th = g->y;

    // Pick the closest larger texture size
    if (tw > 128) {
	tw = 256;
	lod = GR_LOD_256;
    } else if (tw > 64) {
	tw = 128;
	lod = GR_LOD_256;
    } else if (tw > 32) {
	tw = 64;
	lod = GR_LOD_128;
    } else if (tw > 16) {
	tw = 32;
	lod = GR_LOD_64;
    } else if (tw > 8) {
	tw = 16;
	lod = GR_LOD_32;
    } else if (tw > 4) {
	tw = 8;
	lod = GR_LOD_16;
    } else if (tw > 2) {
	tw = 4;
	lod = GR_LOD_4;
    } else if (tw > 1) {
	tw = 2;
	lod = GR_LOD_2;
    } else
	quit("Bad texture width (1 or 0)");

    if (th > 128)		th = 256;
    else if (th > 64)           th = 128;
    else if (th > 32)           th = 64;
    else if (th > 16)		th = 32;
    else if (th > 8)            th = 16;
    else if (th > 4)            th = 8;
    else if (th > 2)            th = 4;
    else if (th > 1)            th = 2;
    else
	quit("Bad texture height (1 or 0)");

    // Calc aspect ratio
    if (tw > th) {
	switch (tw / th) {
	case 1:
	    asp = GR_ASPECT_1x1;
	    break;
	case 2:
	    asp = GR_ASPECT_2x1;
	    break;
	case 4:
	    asp = GR_ASPECT_4x1;
	    break;
	case 8:
	    asp = GR_ASPECT_8x1;
	    break;
	default:
	    asp = -1;
	}
    } else {
	switch (th / tw) {
	case 1:
	    asp = GR_ASPECT_1x1;
	    break;
	case 2:
	    asp = GR_ASPECT_1x2;
	    break;
	case 4:
	    asp = GR_ASPECT_1x4;
	    break;
	case 8:
	    asp = GR_ASPECT_1x8;
	    break;
	default:
	    asp = -1;
	}
    }
    if (asp == -1)
	quit("Unable to determine aspect ratio");

    // only 1 Lod
    g->info.smallLod = lod;
    g->info.largeLod = lod;
    g->info.aspectRatio = asp;
    g->info.format = GR_TEXFMT_P_8;

    textureSize = grTexTextureMemRequired(evenodd, &g->info);
    if (textureSize != (FxU32) tw * th)
	quit("Miscalculated required memory");

    // alloc mem for the scaled texture
    g_info_data = (char *) malloc(tw * th);
    if (!g_info_data)
	quit("Not enough memory for scaled texture");

    // scaling formula.
    for (i = 0; i < th; i++)
	for (j = 0; j < tw; j++) {
	    pixel = g->buffer[(int) ((float) j * g->x / tw + i * g->y / th * g->x)];
	    g_info_data[i * tw + j] = lookup[pixel];
	}

    // point the info.data to the new scaled texture
    g->info.data = g_info_data;

    // put it in texture memory
    g->TextureAddress = getStartAddress(evenodd, &g->info);

    if (g->TextureAddress == -1)
	quit("getStartAddress: out of texture mem "
	     "(handling not implemented yet!)");

    // download the tex
    grTexDownloadMipMap(GR_TMU0,
			g->TextureAddress,
			evenodd,
			&g->info);

    // store the new (scaled) width and height
    g->new_x = tw;
    g->new_y = th;
}

int             ReadthePIC(char *filename, PIC * piccy)
{
    char            infname[132];
    int             j;

    strcpy(infname, filename);
    if (!strchr(infname, '.'))
	strcat(infname, ".gif");

    if (!GetPIC(infname, piccy))
	return 0;

    return 1;
}

void            Pic2FX(char *name, PIC * g)
{
    memset(g, 0, sizeof(PIC));
    if (!ReadthePIC(name, g)) {
	char            er[132];
	sprintf(er, "GetPic failed: %s\n", errtext);
	quit(er);
    }
    PicToTMU(g);

#ifndef BANSHEE
    printf("Picture: %ix%ix%i -> ", g->x, g->y, g->c);
    printf("%ix%i\n", g->new_x, g->new_y);
#endif
}

void            ReadPicTex(int count)
{
    struct find_t   ffblk;
    unsigned        done, j = 0;

    done = _dos_findfirst("*.gif", _A_NORMAL, &ffblk);
    if (done)
	quit("no *.gif found");
    // tex 1
    Pic2FX(ffblk.name, &pics[j]);

    // next tex's
    for (j = 1; j < count; j++)
	if (done)
	    memcpy(&pics[j], &pics[0], sizeof(PIC));
	else {
	    done = _dos_findnext(&ffblk);
	    if (done)
		memcpy(&pics[j], &pics[0], sizeof(PIC));
	    else
		Pic2FX(ffblk.name, &pics[j]);
	}
}

#ifndef BANSHEE
int             CheckDualMonitor()
{
    char           *ptr = getenv("SST_DUALHEAD");
    if (!ptr || ptr[0] == '0') {
	puts("\nOptions:\nPress TAB during viewing to see the options\n");
	delay(1000);
	return NO;
    }
    return YES;
}
#endif
void            main()
{
    int dual;
    puts("\n=== 256-Color Textured Cubes ===\n");

#ifndef BANSHEE
    dual = CheckDualMonitor();
#endif

    GenInit();
    SetupGlideStates();
    MainLoop();

    quit(ALL_FINE);
}
/* End */
