/*
   GIF Loader, reads interlaced too.
   Based upon "Gif Load Source for WGT - Free Code"

   v1.1.1
   - Error type returned in 'errtext'
   - GIF struct added for use with Glide.
   - Tested with Watcom mainly
 */
#include <stdlib.h>
#include <stdio.h>
#include <mem.h>
#include <dos.h>
#include <malloc.h>

#include "usualdef.h"		// favorites (uchar, etc)

char            errtext[132];
char            errcode[8][20] =
{
    "GOOD_READ"
    ,"BAD_FILE"
    ,"BAD_READ"
    ,"UNEXPECTED_EOF"
    ,"BAD_CODE"
    ,"BAD_FIRSTCODE"
    ,"BAD_ALLOC"
    ,"BAD_SYMBOLSIZE"
};

#define GOOD_READ       0
#define BAD_FILE        1
#define BAD_READ        2
#define UNEXPECTED_EOF  3
#define BAD_CODE        4
#define BAD_FIRSTCODE   5
#define BAD_ALLOC       6
#define BAD_SYMBOLSIZE  7

#define NO_CODE         0xFFFF
#define cshort      	cval

#define largest_code    4095
#define table_size      5003

#if __GNUC__
// 2.01 optimized struct's...
#define PADDING   	char	    padding;
#define PADDINGVAL	-1

#else
#define PADDING
#define PADDINGVAL

#endif

#pragma pack(1)

typedef struct {
    unsigned short  x, y, c, i;
    ushort          new_x, new_y;
    GrTexInfo       info;
    unsigned        TextureAddress;
    char            palette[768];
    char           *buffer;
} GIF, PIC;

typedef struct {
    signed char     sig[6];
    unsigned short  screenwidth, screendepth;
    unsigned char   flags, background, aspect;
                    PADDING
} GIFHEADER;

typedef struct {
    unsigned short  left, top, width, depth;
    unsigned char   flags;
                    PADDING
} IMAGEBLOCK;

typedef struct {
    short           width, depth, bits;
    short           flags;
} FILEINFO;

typedef struct {
    signed char     blocksize;
    signed char     flags;
    unsigned short  delay;
    signed char     transparent_colour;
    signed char     terminator;
} CONTROLBLOCK;

typedef struct {
    signed char     blocksize;
    unsigned short  left, top;
    unsigned short  gridwidth, gridheight;
    signed char     cellwidth, cellheight;
    signed char     forecolour, backcolour;
                    PADDING
} PLAINTEXT;

typedef struct {
    signed char     blocksize;
    signed char     applstring[8];
    signed char     authentication[3];
} APPLICATION;

static FILEINFO fi;

#pragma pack()

void            putaline(uchar * p, long n, char *ptr)
{
    if (n >= 0 && n < fi.depth)
	memcpy(ptr + ((long) n * (long) fi.width),
	       p, fi.width);
}

void            putextension(FILE *fp)
{
    PLAINTEXT       pt;
    CONTROLBLOCK    cb;
    APPLICATION     ap;
    short           c, n, i;

    fread(&c, 1, 1, fp);
    switch (c) {
    case 0x0001:
	fread((char *) &pt, 1, sizeof(PLAINTEXT) PADDINGVAL, fp);
	do {
	    if ((n = fgetc(fp)) != EOF) {
		for (i = 0; i < n; i++)
		    fgetc(fp);
	    }
	} while (n > 0 && n != EOF);
	break;
    case 0x0009:
	fread((char *) &cb, 1, sizeof(CONTROLBLOCK), fp);
	break;
    case 0x00fe:
	do {
	    if ((n = fgetc(fp)) != EOF)
		for (i = 0; i < n; i++)
		    fgetc(fp);
	} while (n > 0 && n != EOF);
	break;
    case 0x00ff:
	fread((char *) &ap, 1, sizeof(APPLICATION), fp);
	do {
	    if ((n = fgetc(fp)) != EOF)
		for (i = 0; i < n; i++)
		    fgetc(fp);
	} while (n > 0 && n != EOF);
	break;
    default:
	if ((n = fgetc(fp)) != EOF) {
	    for (i = 0; i < n; i++)
		fgetc(fp);
	}
	break;
    }
}

short           unpackimage(FILE *fp, short bits, GIF * g)
{
    ushort          bits2;
    ushort          codesize;
    ushort          codesize2;
    ushort          nextcode;
    ushort          thiscode;
    ushort          oldtoken;
    ushort          currentcode;
    ushort          oldcode;
    ushort          bitsleft;
    ushort          blocksize;
    ushort          line = 0;
    ushort          byte = 0;
    ushort          pass = 0;

    uchar          *p;
    uchar          *q;
    uchar           b[255];
    uchar          *u;
    uchar          *linebuffer;

    static uchar    firstcodestack[4096];
    static uchar    lastcodestack[4096];
    static ushort   codestack[4096];

    static ushort   wordmasktable[] =
    {
	0x0000, 0x0001, 0x0003, 0x0007,
	0x000f, 0x001f, 0x003f, 0x007f,
	0x00ff, 0x01ff, 0x03ff, 0x07ff,
	0x0fff, 0x1fff, 0x3fff, 0x7fff};

    static ushort   inctable[] =
    {8, 8, 4, 2, 0};
    static ushort   startable[] =
    {0, 4, 2, 1, 0};

    p = q = b;
    bitsleft = 8;

    if (bits < 2 || bits > 8)
	return (BAD_SYMBOLSIZE);
    bits2 = 1 << bits;
    nextcode = bits2 + 2;
    codesize2 = 1 << (codesize = bits + 1);
    oldcode = oldtoken = NO_CODE;

    if ((linebuffer = (char *) malloc(fi.width)) == 0)
	return (BAD_ALLOC);

    /* loop until something breaks */
    for (;;) {
	if (bitsleft == 8) {
	    if (++p >= q && (((blocksize = fgetc(fp)) < 1) ||
	    (q = (p = b) + fread(b, 1, blocksize, fp)) < (b + blocksize))) {
		free(linebuffer);
		return (UNEXPECTED_EOF);
	    }
	    bitsleft = 0;
	}
	thiscode = *p;
	if ((currentcode = (codesize + bitsleft)) <= 8) {
	    *p >>= codesize;
	    bitsleft = currentcode;
	} else {
	    if (++p >= q && (((blocksize = fgetc(fp)) < 1) ||
	    (q = (p = b) + fread(b, 1, blocksize, fp)) < (b + blocksize))) {
		free(linebuffer);
		return (UNEXPECTED_EOF);
	    }
	    thiscode |= *p << (8 - bitsleft);
	    if (currentcode <= 16)
		*p >>= (bitsleft = currentcode - 8);
	    else {
		if (++p >= q && (((blocksize = fgetc(fp)) < 1) ||
				 (q = (p = b) + fread(b, 1, blocksize, fp)) < (b + blocksize))) {
		    free(linebuffer);
		    return (UNEXPECTED_EOF);
		}
		thiscode |= *p << (16 - bitsleft);
		*p >>= (bitsleft = currentcode - 16);
	    }
	}
	thiscode &= wordmasktable[codesize];
	currentcode = thiscode;

	if (thiscode == (bits2 + 1))
	    break;
	if (thiscode > nextcode) {
	    free(linebuffer);
	    return (BAD_CODE);
	}
	if (thiscode == bits2) {
	    nextcode = bits2 + 2;
	    codesize2 = 1 << (codesize = (bits + 1));
	    oldtoken = oldcode = NO_CODE;
	    continue;
	}
	u = firstcodestack;

	if (thiscode == nextcode) {
	    if (oldcode == NO_CODE) {
		free(linebuffer);
		return (BAD_FIRSTCODE);
	    }
	    *u++ = oldtoken;
	    thiscode = oldcode;
	}
	while (thiscode >= bits2) {
	    *u++ = lastcodestack[thiscode];
	    thiscode = codestack[thiscode];
	}

	oldtoken = thiscode;
	do {
	    linebuffer[byte++] = thiscode;
	    if (byte >= fi.width) {
		putaline(linebuffer, line, g->buffer);
		byte = 0;
		/* check for interlaced image */
		if (fi.flags & 0x40) {
		    line += inctable[pass];
		    if (line >= fi.depth)
			line = startable[++pass];
		} else
		    ++line;
	    }
	    if (u <= firstcodestack)
		break;
	    thiscode = *--u;
	} while (1);

	if (nextcode < 4096 && oldcode != NO_CODE) {
	    codestack[nextcode] = oldcode;
	    lastcodestack[nextcode] = oldtoken;
	    if (++nextcode >= codesize2 && codesize < 12)
		codesize2 = 1 << ++codesize;
	}
	oldcode = currentcode;
    }
    free(linebuffer);
    return (GOOD_READ);
}

int             GetPIC(char *filename, GIF * g)
{
    FILE           *filehandle;
    GIFHEADER       gh;
    IMAGEBLOCK      iblk;
    long            t;
    short           b;
    unsigned short  cval;
    uchar           cbyte;
    uchar           cbytes[5];
    char            file[132];

    g->i = 0;
    strcpy(errtext, "No error (init)");
    strcpy(file, filename);
    if (strchr(file, '.') == 0)
	strcat(file, ".gif");

    if ((filehandle = fopen(file, "rb")) == 0) {
	strcpy(errtext, "File not found");
	return 0;
    }
#ifndef BANSHEE
    printf("Loading %s...\n", file);
#endif

    if (fread((char *) &gh, 1, sizeof(GIFHEADER) PADDINGVAL, filehandle) != sizeof(GIFHEADER) PADDINGVAL
	|| memcmp(gh.sig, "GIF", 3)) {
	fclose(filehandle);
	strcpy(errtext, "GIF header not found");
	return 0;
    }
    g->x = fi.width = gh.screenwidth;
    g->y = fi.depth = gh.screendepth;
    g->c = fi.bits = (gh.flags & 0x0007) + 1;
    g->c = (1 << g->c);

    if (gh.flags & 0x80) {
	cshort = 3 * (1 << ((gh.flags & 7) + 1));
	if (fread(g->palette, 1, cshort, filehandle) != cshort) {
	    fclose(filehandle);
	    strcpy(errtext, "Color map not found");
	    return 0;
	}
    }
    while ((cbyte = fgetc(filehandle)) == ',' || cbyte == '!' || cbyte == 0) {
	cval = cbyte;
	if (cval == ',') {
	    if (fread(&iblk, 1, sizeof(IMAGEBLOCK) PADDINGVAL, filehandle) != sizeof(IMAGEBLOCK) PADDINGVAL) {
		fclose(filehandle);
		strcpy(errtext, "start of the image block not found");
		return 0;
	    }
	    fi.width = iblk.width;
	    fi.depth = iblk.depth;

	    if (iblk.flags & 0x80) {
		b = 3 * (1 << ((iblk.flags & 0x0007) + 1));
		if (fread(g->palette, 1, b, filehandle) != cshort) {
		    fclose(filehandle);
		    strcpy(errtext, "local colour map not found");
		    return 0;
		}
		fi.bits = (iblk.flags & 0x0007) + 1;
	    }
	    if (fread(&cshort, 1, 1, filehandle) == 0) {
		fclose(filehandle);
		strcpy(errtext, "initial code size not found");
		return 0;
	    }
	    fi.flags = iblk.flags;
	    if (fi.flags & 0x40)
		g->i++;

	    if ((g->buffer = (char *) malloc((long) fi.width * (long) fi.depth + 5)) == 0) {
		fclose(filehandle);
		strcpy(errtext, "no mem");
		return 0;
	    }
	    t = unpackimage(filehandle, cshort, g);
	    if (t != GOOD_READ) {
		free(g->buffer);
		fclose(filehandle);
		sprintf(errtext, "unpack failed [%s]", errcode[t]);
		return 0;
	    } else {
		fclose(filehandle);
		strcpy(errtext, "Done.");
		return 1;
	    }
	} else if (cval == '!')
	    putextension(filehandle);

    }
    fclose(filehandle);
    strcpy(errtext, "All done.");
    return 1;
}
