/*
   Usual Glide & Program Stuff
   (code that changes less often.)
   v0.5.1
 */

FxU32           textureSize, nextTexture, lastTexture;

void            StartGlideExtras();

/*
   Snap functions
 */

void            lowfpu()
{
    long            memvar;
    _asm {
	finit			// initialize the FPU
	     fwait		// wait for operation to complete
	     fstcw[memvar]	// store FPU control word to memvar
	     fwait		// wait for operation to complete
	     mov eax,[memvar]	// move memvar to a register
	     and            eax, 0f ffffcffh	// mask off precision bits to set to 24-bit precision
	                mov[memvar], eax	// save control word to memory
	                fldcw[memvar]	// load control word back to FPU
	                fwait	// wait for operation to complete
    }
}

#define SNAPPER (float)(3L << 18)
#define SNAP(vertex) \
(vertex).x += SNAPPER; (vertex).x -= SNAPPER; \
(vertex).y += SNAPPER; (vertex).y -= SNAPPER;

/*
   exit routine
 */
void        quit(char *t)
{
    if (GlideIsRunning)
	grGlideShutdown();
    GlideIsRunning = 0;
    if (*t) {
	printf("\n=== Error: %s!\n", t);
	exit(1);
    }
    puts("\nBye.");
    exit(0);
}

/*
   Matrix cube.
   S1-S2-S3 are adjacent
 */
void            BuildCube(MAT3D S[])
{
    MAT3D           Flip, Temp;
    int             j = 0;

    Flip = RotateY(180. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    Flip = Temp * Flip;

    S[j] = Translate(0, 0, 0);

    j++;
    Temp = RotateX(-90. * 3.141592654 / 180.);
    S[j] = Temp * Flip;

    j++;
    Temp = Translate(side2 - side1, 0, 0);
    S[j] = Temp * Flip;
    Temp = RotateY(90. * 3.141592654 / 180.);
    S[j] = Temp * S[j];

    j++;
    Temp = Translate(0, 0, side2);
    S[j] = Temp * Flip;

    j++;
    S[j] = RotateX(-90. * 3.141592654 / 180.);
    Temp = Translate(0, side2, 0);
    S[j] = Temp * S[j];

    j++;
    S[j] = RotateY(90. * 3.141592654 / 180.);
    Temp = Translate(side1, 0, 0);
    S[j] = Temp * S[j];
}

/*
   Inits
 */

void            StartGlide()
{
    grGlideInit();
    if (!(grSstQueryHardware(&hwconfig)))
	quit("Query Hardware failed");
    grSstSelect(0);
    if (!(grSstWinOpen(0,
		       GR_RESOLUTION_640x480,
		       GR_REFRESH_60Hz,
		       GR_COLORFORMAT_ABGR,
		       GR_ORIGIN_LOWER_LEFT,	//faces outside
    //                      GR_ORIGIN_UPPER_LEFT, //faces inside
		        2,
		       1
	  )))
	quit("Win Open failed");
    GlideIsRunning = YES;
}

/*
   Texture routines
 */

long            getStartAddress(FxU32 evenOdd, GrTexInfo * info)
{
    long            start;
    static int      Totalsize = 0;
    static short    initboundsflag = 0;

    if (!initboundsflag) {
	nextTexture = grTexMinAddress(GR_TMU0);
	lastTexture = grTexMaxAddress(GR_TMU0);
	initboundsflag = 1;
    }
    //sum of the sizes of the even/odd/all LODs
    textureSize = grTexTextureMemRequired(evenOdd, info);

    Totalsize += textureSize;
#ifndef BANSHEE
    printf("Total texture memory used = %i (%ik)\n", Totalsize, Totalsize / 1024);
#endif

    start = nextTexture;
    nextTexture += textureSize;
    if (nextTexture <= lastTexture)
	return start;

    nextTexture = start;

    return -1;
}

/*---*/
